/*******************************************************************************
 *     ___                  _   ____  ____
 *    / _ \ _   _  ___  ___| |_|  _ \| __ )
 *   | | | | | | |/ _ \/ __| __| | | |  _ \
 *   | |_| | |_| |  __/\__ \ |_| |_| | |_) |
 *    \__\_\\__,_|\___||___/\__|____/|____/
 *
 *  Copyright (c) 2014-2019 Appsicle
 *  Copyright (c) 2019-2022 QuestDB
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 ******************************************************************************/

package io.questdb.cutlass.line.udp;

import io.questdb.cairo.*;
import io.questdb.cairo.pool.PoolListener;
import io.questdb.cutlass.line.AbstractLineSender;
import io.questdb.cutlass.line.LineUdpSender;
import io.questdb.mp.SOCountDownLatch;
import io.questdb.network.Net;
import io.questdb.network.NetworkFacadeImpl;
import io.questdb.std.Chars;
import io.questdb.std.Os;
import io.questdb.test.tools.TestUtils;
import org.junit.Assert;

import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

public abstract class LineUdpInsertTest extends AbstractCairoTest {

    protected static final int LOCALHOST = Net.parseIPv4("127.0.0.1");
    protected static final LineUdpReceiverConfiguration RCVR_CONF = new DefaultLineUdpReceiverConfiguration();
    protected static final int PORT = RCVR_CONF.getPort();

    protected static void assertReader(String tableName, String expected) {
        assertReader(tableName, expected, (String[]) null);
    }

    protected static void assertReader(String tableName, String expected, String... expectedExtraStringColumns) {
        try (TableReader reader = new TableReader(new DefaultCairoConfiguration(root), tableName)) {
            TestUtils.assertReader(expected, reader, sink);
            if (expectedExtraStringColumns != null) {
                TableReaderMetadata meta = reader.getMetadata();
                Assert.assertEquals(2 + expectedExtraStringColumns.length, meta.getColumnCount());
                for (String colName : expectedExtraStringColumns) {
                    Assert.assertEquals(ColumnType.STRING, meta.getColumnType(colName));
                }
            }
        }
    }

    protected static void assertType(String tableName,
                                     String targetColumnName,
                                     int columnType,
                                     String expected,
                                     Consumer<AbstractLineSender> senderConsumer,
                                     String... expectedExtraStringColumns) throws Exception {
        TestUtils.assertMemoryLeak(() -> {
            try (CairoEngine engine = new CairoEngine(configuration, metrics, 2)) {
                final SOCountDownLatch waitForData = new SOCountDownLatch(1);
                engine.setPoolListener((factoryType, thread, name, event, segment, position) -> {
                    if (event == PoolListener.EV_RETURN && Chars.equals(tableName, name)) {
                        waitForData.countDown();
                    }
                });
                try (AbstractLineProtoUdpReceiver receiver = createLineProtoReceiver(engine)) {
                    if (columnType != ColumnType.UNDEFINED) {
                        try (TableModel model = new TableModel(configuration, tableName, PartitionBy.NONE)) {
                            CairoTestUtils.create(model.col(targetColumnName, columnType).timestamp());
                        }
                    }
                    receiver.start();
                    try (AbstractLineSender sender = createLineProtoSender()) {
                        sender.disableValidation();
                        senderConsumer.accept(sender);
                        sender.flush();
                    }
                    Os.sleep(250L); // allow reader to hit the readout
                }

                if (!waitForData.await(TimeUnit.SECONDS.toNanos(30L))) {
                    Assert.fail();
                }

                assertReader(tableName, expected, expectedExtraStringColumns);
            }
        });
    }

    protected static AbstractLineProtoUdpReceiver createLineProtoReceiver(CairoEngine engine) {
        AbstractLineProtoUdpReceiver lpr;
        if (Os.type == Os.LINUX_AMD64) {
            lpr = new LinuxMMLineUdpReceiver(RCVR_CONF, engine, null);
        } else {
            lpr = new LineUdpReceiver(RCVR_CONF, engine, null);
        }
        return lpr;
    }

    protected static AbstractLineSender createLineProtoSender() {
        return new LineUdpSender(NetworkFacadeImpl.INSTANCE, 0, LOCALHOST, PORT, 80, 1);
    }
}